#pragma once
#include <vrj/vrjConfig.h>

#include <iostream>
#include <iomanip>

#include <gmtl/Matrix.h>
#include <gmtl/MatrixOps.h>
#include <vrj/Draw/OGL/GlApp.h>

#include <gadget/Type/PositionInterface.h>
#include <gadget/Type/DigitalInterface.h>
#include <gadget/Type/AnalogInterface.h>
#include <gadget/Type/KeyboardMouseInterface.h>

#include <vrj/Kernel/User.h>

using namespace vrj;

class Universe;
class Controller;

class Application : public GlApp
{
public:
	Application(Universe *uni);
	virtual ~Application(void);

	//init methods
	virtual void init();
    virtual void apiInit();
    virtual void contextInit();

	//drawing loop methods
	virtual void bufferPreDraw();
    virtual void preFrame();
    virtual void intraFrame();
    virtual void postFrame();
    virtual void draw();

protected:
	Universe *universe;
	Controller *controller;

	//VR Juggler input data
	gadget::PositionInterface mWand; /**< Positional interface for Wand position */
	gadget::PositionInterface mHead; /**< Positional interface for Head position */
	gadget::KeyboardMouseInterface keyboard;
    gadget::DigitalInterface   mIncVelocityButton;     // Button for velocity
    gadget::DigitalInterface   mDecVelocityButton;
    gadget::DigitalInterface   mStopButton;            // Button to stop
    gadget::AnalogInterface    mJoystick[2];           // 2D Joystick

	// Navigation info for the user
    float                mCurVelocity;  // The current velocity
    gmtl::Matrix44f      nav_mat;    // Matrix for navigation in the application

    vrj::User*           mUser;         // The user we hold data for

    // Defaults for joystick navigation
    float mSpeedMax;
    float mSpeedExponent;
    float mRotateSpeedMax;
    float mRotateSpeedExponent;

	//cal 3d variables
	unsigned int m_lastTick;
	int countdown;
	bool m_bPaused;

};
