#include "StdAfx.h"
#include "polygon.h"
#include "ShadowVolume.h"
#ifdef VC71
#include <GL/glut.h>
#else
#include "glut.h"
#endif
#include <iostream>


using namespace std;

extern Vertex lightPos;

//p is the polyhedran that this polygon is part of (needs the reference for the vertex list)
Polygon::Polygon()
{
	normal = NULL;
}

Polygon::~Polygon(void)
{
	for (unsigned int j = 0; j < vertexList.size(); ++j)
	{
		delete vertexList[j];
		delete texList[j];
	}
	delete normal;
}

//renders this polygon using OpenGL
void Polygon::render()
{
	glBegin(GL_POLYGON);

	//output this polygon
	Vertex normal = *getNormal();
	glNormal3f(normal(0), normal(1), normal(2));
	for (unsigned int i = 0; i < vertexList.size(); ++i)
	{
		Vertex vert = *getVertex(i);
		Vertex tex = *texList[i];

		glTexCoord2f(tex(0), tex(1));
		glVertex3f(vert(0), vert(1), vert(2));
		//cout << tex(0) << ", " << tex(1) << "\n";
		//cout << vert(0) << ", " << vert(1) << ", " << vert(2) << "\n";
	}
	glEnd();

	//draw the normals for debugging
	/*glBegin(GL_LINES);
	for (unsigned int i = 0; i < vertexList.size(); ++i)
	{
		Vertex a = *getVertex(i);
		Vertex b = *getVertex(i) + normal;

		glVertex3f(a(0), a(1), a(2));
		glVertex3f(b(0), b(1), b(2));
	}
	glEnd();*/
}

Vertex Polygon::calcNormal(Vertex &v0, Vertex &v1, Vertex &v2, Vertex &v3)
{
	//two vectors in the polygon
	Vertex vA = v0 - v2;
	Vertex vB = v1 - v3;

	Vertex result = crossProduct(vA, vB);

	//normalise the cross product
	result.normalise();
	return result;
}

void Polygon::calculatePlaneEqn(Vertex &v0, Vertex &v1, Vertex &v2, double *result)
{
	Vertex normal = calcNormal(v2, v1, v0, v0);
	result[3] = dotProduct(normal, v0);

	result[0] = normal(0);
	result[1] = normal(1);
	result[2] = normal(2);

	delete &normal;

	//cout << result[0] << ' ' << result[1] << ' ' << result[2] << ' ' << result[3] << '\n';
}

//calculate the normal vector of this polygon by calculating the cross product of two
//vectors within it
Vertex *Polygon::getNormal(void)
{
	if (normal == NULL)
	{
		if (vertexList.size() <= 2)
			normal = new Vertex(0, 0, 0);
		else
			normal = new Vertex(calcNormal(*(getVertex(0)), *(getVertex(1)), *(getVertex(2)), *(getVertex(3))));
	}

	return normal;
}


Vertex Polygon::getCenter()
{
	Vertex result;
	//sum all the vertices
	for (unsigned int i = 0; i < vertexList.size(); ++i)
	{
		Vertex curVert = *getVertex(i);
		result = result + curVert;
	}

	//find the average of the vertices
	result(0) = result(0) / vertexList.size();
	result(1) = result(1) / vertexList.size();
	result(2) = result(2) / vertexList.size();

	return result;
}

void Polygon::renderShadowVolumes(Vertex *lightPosition)
{
	//cout << (*vertexList[0])(0) << ", " << (*vertexList[0])(1) << ", " << (*vertexList[0])(2) << "\n";
	//cout << (*vertexList[1])(0) << ", " << (*vertexList[1])(1) << ", " << (*vertexList[1])(2) << "\n";
	//cout << (*vertexList[2])(0) << ", " << (*vertexList[2])(1) << ", " << (*vertexList[2])(2) << "\n";
	//cout << (*vertexList[3])(0) << ", " << (*vertexList[3])(1) << ", " << (*vertexList[3])(2) << "\n";
	ShadowVolume::renderShadowVolumesDirect(this, lightPosition);
}


void Polygon::rotateY(float rads)
{
	//rotate all the vertices of the polygon
	for (unsigned int i = 0; i < vertexList.size(); ++i)
	{
		vertexList[i]->rotateY(rads);
	}
}
