#include "StdAfx.h"
#include "texture.h"
#include <ios>
#ifdef VC71
#include <fstream>
using namespace std;
#else
#include <fstream.h>
#endif

#ifdef VC71
#include <GL/glut.h>
#else
#include "glut.h"
#endif

//using namespace std;

Texture::Texture(char *fileName)
{
	//load the image from file
	loadImage(fileName);

	//save a copy of the filename
	fName = (char *) malloc(strlen(fileName)+1);
	strcpy(fName, fileName);
}

void Texture::loadToGL(void)
{
	if (image != 0)
	{
		//get the next texture id from opengl
		glGenTextures(1, &(*textureId));

		//load the image to opengl texture mem
		loadTexture();
	}
}


Texture::~Texture(void)
{
	delete [] image;
	free(fName);
}

void Texture::loadImage(char *fileName)
{
	//cout << fileName << '\n';

	//loader for my own simple image file
	//short header of 8 bytes followed by the image in such a format that it can
	//be sent direct to opengl
	//I use a small java app to convert images to this format
	//file format
	//------------
	//byte	use
	//------------
	//0		a magic number to id the family of file types, should = 26
	//1		specific file type within family, should = 1
	//2		version number of file type, should = 1
	//3		padding
	//4, 5	width
	//6, 7	height
	//8+	the image, interlaced
	typedef unsigned char byte;

	ifstream in(fileName, ifstream::binary);
	in.unsetf(std::ios_base::skipws);

	if (!in.is_open())
	{
		cout << "Error: could not open file" << '\n';
		width = height = 0;
		image = 0;
		in.close();
		return;
	}

	byte magicNo, fileType, versionNo, trans;
	byte widthMSB, widthLSB, heightMSB, heightLSB;

	in >> magicNo;
	in >> fileType;
	in >> versionNo;

	if (magicNo != 26 || fileType != 1 || versionNo != 1)
	{
		cout << "Error: bad file type " << '\n';
		cout << magicNo << ' ' << fileType << ' ' << versionNo << '\n';
		width = height = 0;
		image = 0;
		in.close();
		return;
	}

	in >> trans;

	in >> widthMSB;
	in >> widthLSB;
	in >> heightMSB;
	in >> heightLSB;

	width = (widthMSB << 8) + widthLSB;
	height = (heightMSB << 8) + heightLSB;

	image = new byte[width * height * 3];

	byte temp;
	for (unsigned int i = 0; i < width*height*3; ++i)
	{
		in >> temp;

		image[i] = temp;
	}

	in.close();
}

void Texture::loadTexture(void)
{
	//identify the texture
	glBindTexture(GL_TEXTURE_2D, *textureId);
	//loading parameters
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	//how the texture interacts with its environment
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	//load the texture
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, image);
}

void Texture::render(void)
{
	//set the colour to white and bind this texture for use
	GLfloat colour[] = {1.0, 1.0, 1.0, 1.0};
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colour);
	GLfloat black[] = {0.0, 0.0, 0.0, 1.0};
	glMaterialfv(GL_FRONT, GL_SPECULAR, black);

	//cout << textureId << '\n';
	//cout.flush();
	glBindTexture(GL_TEXTURE_2D, *textureId);
}
