#pragma once
#include <iostream>
#include "StdAfx.h"
#include "user.h"
#include "Camera.h"
#include "Universe.h"
#include "Box2.h"
#include "Cal3DAvatar.h"
#include "Vertex.h"
#include "Light.h"
#ifdef VC71
#include <GL/glut.h>
#else
#include "glut.h"
#endif

#define _USE_MATH_DEFINES 
#include <math.h>

using namespace std;

extern bool linkUser;

User::User(Universe *u) : yMouse(0)
{
	camera = new Camera();
	owner = u;

	state = walking;
}

User::~User(void)
{
	delete camera;
	delete rend;
}

void User::render(void)
{
	//we need to transform the avatar so it is located and aligned correctly
	glPushMatrix();
	glTranslatef(positionX, 0, positionZ);
	glRotatef(180-degreesDir, 0.0, 1.0, 0.0);
	rend->render();
	glPopMatrix();
}

void User::update(float elapsedSeconds)
{
	//update avatar and states
	if (state == walking)
		rend->update(elapsedSeconds);

	state = resting;
}

void User::enterWalkingState(void)
{
	state = walking;
}

void User::renderShadowVolumes(Vertex *lightPosition)
{
	//we need to take into account possible rotation and
	//translation of the avatar
	glPushMatrix();
	glTranslatef(positionX, 0, positionZ);
	glRotatef(-degreesDir, 0.0, 1.0, 0.0);

	//adjust the light so it is in the right place
	Vertex translatedLightPos = Vertex(*lightPosition);
	translatedLightPos.translate(-positionX, 0, -positionZ);
	translatedLightPos.rotateY(radiansDir);

	//rend->renderShadowVolumes(&translatedLightPos);

	glPopMatrix();
}

void User::renderProjectedShadows(Vertex *lightPosition)
{
	//render the Renderable's shadow in the correct place
	glPushMatrix();
	glTranslatef(positionX, 0, positionZ);
	glRotatef(180-degreesDir, 0.0, 1.0, 0.0);

	//adjust the light so it is in the right place
	Vertex translatedLightPos = Vertex(*lightPosition);
	translatedLightPos.translate(-positionX, -positionY, -positionZ);
	translatedLightPos.rotateY(M_PI+radiansDir);

	rend->renderProjectedShadows(&translatedLightPos, 0);

	glPopMatrix();
}

void User::init(void)
{
	//rend = Box2::getInstance();
	rend = new Cal3DAvatar();

	camera->init();
	rend->init();

	positionZ = -2;
	positionX = 10;
	positionY = 1;

	//use the velocity as a direction vector
	velocityY = 0;

	radiansDir = 4.7;
	degreesDir = radiansDir * 180 / M_PI;
	velocityX = -sin(-radiansDir);
	velocityZ = -cos(-radiansDir);

	camera->setPosition(positionX, 1, positionZ);
	camera->rotate(degreesDir);
}

void User::onKeyPress(unsigned char c, int x, int y)
{
//walking speed
#define MULTIPLIER (0.085)

	//adjust the state and position of the avatar
	//this is a bit of a hack; we should just set the velocity and state
	//and use update(time) to change the position according to the velocity
	if (c == 'w')
	{
		if (linkUser)
		{
			positionZ += velocityZ * MULTIPLIER;
			positionX += velocityX * MULTIPLIER;
		}
		enterWalkingState();
		camera->translate(velocityX * MULTIPLIER, 0, velocityZ * MULTIPLIER);
		owner->collide(this);
	}
	else if (c == 's')
	{
		if (linkUser)
		{
			positionZ -= velocityZ * MULTIPLIER;
			positionX -= velocityX * MULTIPLIER;
		}
		enterWalkingState();
		camera->translate(-velocityX * MULTIPLIER, 0, -velocityZ * MULTIPLIER);
		owner->collide(this);
	}
	else if (c == 'a')
	{
		if (linkUser)
		{
			positionZ -= velocityX * MULTIPLIER;
			positionX += velocityZ * MULTIPLIER;
		}
		camera->translate(velocityZ * MULTIPLIER, 0, -velocityX * MULTIPLIER);
		owner->collide(this);
	}
	else if (c == 'd')
	{
		if (linkUser)
		{
			positionZ += velocityX * MULTIPLIER;
			positionX -= velocityZ * MULTIPLIER;
		}
		camera->translate(-velocityZ * MULTIPLIER, 0, velocityX * MULTIPLIER);
		owner->collide(this);
	}
	else if (c == 'r')
	{
		if (!linkUser)
		{
			camera->translate(0, 1, 0);
		}
	}
	else if (c == 'f')
	{
		if (!linkUser)
		{
			camera->translate(0, -1, 0);
		}
	}
}

void User::onMouseMove(int x, int y)
{
	int dx = x - yMouse;

	radiansDir += dx / 50.0;
	degreesDir = radiansDir * 180 / M_PI;
	velocityX = -sin(-radiansDir);
	velocityZ = -cos(-radiansDir);

	camera->rotate(degreesDir);

	yMouse = x;
}

void User::onMouseButton(int button, int state, int x, int y)
{
	//when the user presses a button, record the start position
	if (state == GLUT_DOWN)
	{
		yMouse = x;
	}

	//when the user releases a mouse button rotate the object
	if (state == GLUT_UP)
	{
		onMouseMove(x, y);
	}
}

void User::setX(float x)
{
	camera->translate(x - positionX, 0, 0);
	positionX = x;
}
void User::setY(float y)
{
	camera->translate(0, y - positionY, 0);
	positionY = y;
}
void User::setZ(float z)
{
	camera->translate(0, 0, z - positionZ);
	positionZ = z;
}
