#pragma once
#include "Thing.h"
#include "Controller.h"

class Camera;
class Universe;

//abstraction of the user in the simulation, handles the user's graphical representation
//(avatar), the user's abstract representation in the scene (as  a Thing) and the
//control of the scene by the user (via the camera)

class User : public Thing, public Controller
{
public:
	User(Universe *u);
	virtual ~User(void);

	//rendering funcs from Thing
	void render(void);
	void update(float elapsedSeconds);
	void renderShadowVolumes(Vertex *lightPosition);
	void renderProjectedShadows(Vertex *lightPosition);
	void init(void);

	//Controller funcs
	void onKeyPress(unsigned char c, int x, int y);
	void onMouseMove(int x, int y);
	void onMouseButton(int button, int state, int x, int y);

	//return the camera that is the user's view of the scene
	Camera *getCamera() { return camera; }

	//set the users position in the scene
	void setX(float x);
	void setY(float y);
	void setZ(float z);

private:
	//the camera that is the user's view of the scene
	Camera *camera;

	//the universe of which this user is a part
	Universe *owner;

	//this variable for mouse movement
	int yMouse;

	//the direction the user is facing
	float radiansDir;
	float degreesDir;

	//handle the different states of the user (which map to different animation cycles)
	void enterWalkingState(void);
	enum State {walking, resting};
	State state;

};
