#pragma once
#include "stdafx.h"

using namespace std;

//a point in space (or a vector)
class Vertex
{
public:
	Vertex();
	//usual constructor
	Vertex(double x, double y, double z);
	//copy constructor
	Vertex(const Vertex & v)
	{
		storage[0] = v.storage[0];
		storage[1] = v.storage[1];
		storage[2] = v.storage[2];
	}
	//destructor
	~Vertex(void)
	{
	}

	//return a reference to part of the vertex 0<=i<3
	//x = 0, y = 1, z = 2
	double& operator() (int i)
	{
		assert(i >= 0);
		assert(i < 3);
		return storage[i];
	}
	double operator() (int i) const  //const version
	{
		assert(i >= 0);
		assert(i < 3);
		return storage[i];
	}

	//subtract one vertex from another to make a vector (represented as a Vertex)
	const Vertex operator - (Vertex & other)
	{
		return Vertex(storage[0] - other(0), storage[1] - other(1), storage[2] - other(2));
	}

	//add two vectors to get another (all represented as Vertexs)
	const Vertex operator + (Vertex & other)
	{
		return Vertex(storage[0] + other(0), storage[1] + other(1), storage[2] + other(2));
	}

	//these methods are more efficent and convenient than matrix multiplication
	//for simple transformations
	void translate(double x, double y, double z);
	void scale(double x, double y, double z);

	void rotateY(float rads);

	//returns the modulus of this vector |a|
	double size(void);

	//normalises this vector ie changes a to a / |a|
	void normalise(void);

private:
	//(0, 1, 2, 3)
	double storage[4];
};

//cross product of two vectors
Vertex crossProduct(Vertex &v1, Vertex &v2);
//dot product of two vectors
double dotProduct(Vertex &v1, Vertex &v2);